(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Frdric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2008,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: code.ml,v 1.57 2008/02/11 23:05:59 weis Exp $ *)

(** {3 The code we generate.} *)

open Parsetree
open Longident
open Asttypes
open Printf
open Pr_otype
;;

type lident = Longident.t
;;
(** Lowercase ident. *)
type uident = Longident.t
;;
(** Uppercase ident. *)
type iident = Longident.t
;;
(** Infix ident. *)
type ident = Longident.t
;;
(** Any kind of ident. *)

type iname = string
;;
(** The name of an infix symbole. *)
type tname = string
;;
(** The name of a type declared in the AST. *)

type exp =
   | Constant of ident
   | Var of lident
   | Match of exp * clause list
   | Try of exp * clause list
   | Generator of uident * exp
   | Apply of exp * exp list
   | Infix of exp * iident * exp
   | If of exp * exp * exp
   | Let of lident * pattern list * exp * exp
   | Letrec of lident * pattern list * exp * exp
   | Sequence of exp * exp * exp list
   | Record of (lident * exp) list
   | Tuple of exp list
   | List of exp list
   | Ocaml_pattern of Parsetree.pattern
   | Ocaml_expression of Parsetree.expression
   | Function of clause list
   | Comment of string

and pattern = exp

and clause =
  | Clause of pattern * exp
  | Clause_when of pattern * exp * exp
;;

type structure_item =
   (** Evaluates an expression *)
   | Pstr_eval of exp
   (** Defines a set of possibly recursive functional values *)
   | Pstr_value of Asttypes.rec_flag * (lident * exp list * exp) list
   (** Defines one primitive *)
   | Pstr_primitive of string * Parsetree.value_description
   (** Defines one type *)
   | Pstr_type of (string * Parsetree.type_declaration) list
   (** Opens a module *)
   | Pstr_open of uident
;;

type structure = structure_item list
;;

let topletrec def = Pstr_value (Recursive, def)
;;

(* To add a prefix to an ident. *)
let rec prefix_longident s = function
  | Lident id -> Lident (s ^ id)
  | Ldot (modid, id) -> Ldot (modid, s ^ id)
  | Lapply (modid1, modid2) -> Lapply (modid1, prefix_longident s modid2)
;;

(* Globally predefined identifiers. *)

let make_Constant s = Constant (Lident s)
;;

let underscore = Constant (Lident "_")
;;

let make_Var s = Var (Lident ("moca_" ^ s))
;;

let x = make_Var "x"
and y = make_Var "y"
and z = make_Var "z"
and t = make_Var "t"
and u = make_Var "u"
;;

(* Replace in a pattern all variables by underscore. *)

let rec underscored_ocaml_pattern p =
  { p with ppat_desc = underscored_ocaml_patdesc p.ppat_desc }

and underscored_ocaml_pattern_opt = function
  | None -> None
  | Some p -> Some (underscored_ocaml_pattern p)

and underscored_ocaml_patdesc = function
  | Ppat_var _ -> Ppat_any
  | Ppat_alias (p, _) ->
      Ppat_parens (underscored_ocaml_pattern p)
  | Ppat_tuple l ->
      Ppat_tuple (List.map underscored_ocaml_pattern l)
  | Ppat_construct (lid, popt, b) ->
      Ppat_construct (lid, underscored_ocaml_pattern_opt popt, b)
  | Ppat_variant (lb, popt) ->
      Ppat_variant (lb, underscored_ocaml_pattern_opt popt)
  | Ppat_record l ->
      Ppat_record
        (List.map (fun (lid, p) -> (lid, underscored_ocaml_pattern p)) l)
  | Ppat_array l ->
      Ppat_array (List.map underscored_ocaml_pattern l)
  | Ppat_or (p1, p2 ) ->
      Ppat_or (underscored_ocaml_pattern p1, underscored_ocaml_pattern p2)
  | Ppat_constraint (p, ct) ->
      Ppat_constraint (underscored_ocaml_pattern p, ct)
  | Ppat_parens p ->
      Ppat_parens (underscored_ocaml_pattern p)
  | (Ppat_type _
  | Ppat_any
  | Ppat_constant _) as p -> p
;;

let is_lowercase x = Char.lowercase x = x;;

let rec underscored_pattern = function
  | Var _ -> underscore
  | Tuple l -> Tuple (List.map underscored_pattern l)
  | Generator (uid, exp) -> Generator (uid, underscored_pattern exp)
  | Ocaml_pattern p -> Ocaml_pattern (underscored_ocaml_pattern p)
  | Record l -> Record (List.map (fun (lid, e) -> (lid, underscored_pattern e)) l)
  (* identity on other expressions *)
  | (Ocaml_expression _ | Match _ | Try _ | Apply _
  | Infix _ | If _ | Let _ | Letrec _ | Sequence _
  | Function _ | Comment _ | Constant _ | List _) as x -> x
;;

(* Replace in a non-when clause lhs pattern all variables by underscore. *)
let underscored_clause = function
  | Clause (l, r) -> Clause (underscored_pattern l, r)
  | Clause_when _ as x -> x
;;

(* [genr_args root n] returns the list of strings [root1; ...; rootn].

   Beware: do not modify the code to generate a mere (admittedly
   simpler) "x", since arguments generated by [genr_args] should not be
   "x" or "u" or "t" or "y" or "z": those variables are already used in
   the generating functions as ``global'' (i.e. toplevel) arguments of
   generated functions.  If genr_args would produce "x" for instance,
   we would have a name capture between a pattern of a clause and the
   argument of the generated functions.  To prevent this phenomenon,
   genr_args always adds an integer after the root prefix. Be cautious
   to preserve this invariant! *)

let genr_args root =
  let rec loop accu = function
    | 0 -> accu
    | n ->
	let e = make_Var (root ^ string_of_int n) in
	  loop (e :: accu) (n - 1) in
    loop []
;;

(* [genr_numbered_args root n]
   returns the list of strings [1, make_Var root1; ...; n, make_Var rootn].
   Beware: same as the one for [genr_args]. *)

let genr_numbered_args root n =
  let rec loop accu = function
    | 0 -> accu
    | n ->
	let e = make_Var (root ^ string_of_int n) in
	  loop ((n, e) :: accu) (n - 1) in
  loop [] n
;;

(* Application macros. *)

let apply f xs = Apply (Constant f, xs)
;;

let apply1 f x = Apply (Constant f, [x])
and apply2 f x y = Apply (Constant f, [x; y])
and apply3 f x y z = Apply (Constant f, [x; y; z])
;;

let infix_apply x f y = Infix (x, f, y)
and infix_name x f y = Infix (x, Lident f, y)
;;

(* Other useful auxilliaries. *)

let let0 x v b = Let (x, [], v, b)
and let1 f x v b = Let (f, [x], v, b)
and pair x y = Tuple [x; y]
and raise_error s = apply1 (Lident "raise") (make_Constant s)
;;

(* Printing code. *)

let iter f b = List.iter (f b)
;;

(* To be suppressed. *)
let longident = Pr_otype.blongident
;;

let longidents = prefix_list " " longident
;;

let infix_ident ob s = bprintf ob "%s" s;;
let rec infix_longident ob = function
  | Lident s -> infix_ident ob s
  | Ldot (t, s) -> bprintf ob "%a.%a" infix_longident t infix_ident s
  | Lapply (t1, t2) ->
    bprintf ob "%a(%a)" infix_longident t1 infix_longident t2
;;

let rec needs_parens = function
  | Constant _
  | Var _
  | Generator _
  | Record _
  | Tuple _
  | List _
  | Comment _
  | Ocaml_expression _
  | Apply (_, []) -> false
  | Apply (_, _::_)
  | Ocaml_pattern _
  | Match _
  | Try _
  | Infix _
  | If _
  | Let _
  | Letrec _
  | Sequence _
  | Function _ -> true

let rec exp b = function
  | Constant lid -> bprintf b "%a" longident lid
  | Var lid -> bprintf b "%a" longident lid
  | Ocaml_pattern p -> Pr_otype.pattern b p
  | Ocaml_expression e -> Pr_otype.expression b e
  | Match (e, l) -> bprintf b "\n(match %a with%a)\n" exp e clauses l
  | Try (e, l) -> bprintf b "\n(try %a with%a)\n" exp e clauses l
  | Generator (gid, e) -> bprintf b "(%a %a)" longident gid exp e
  | Apply (e, l) -> papply " " b e l
  | Infix (e1, lid, e2) ->
      bprintf b "(%a %a %a)" exp e1 infix_longident lid exp e2
  | If (e1, e2, e3) ->
    bprintf b "if %a\nthen %a\nelse %a\n" exp e1 exp e2 exp e3
  | Let (fname, args, e1, e2) ->
    bprintf b "\nlet %a%a = %a in %a"
      longident fname patterns args exp e1 exp e2
  | Letrec (fname, args, e1, e2) ->
    bprintf b "\nlet rec %a%a = %a in %a"
      longident fname patterns args exp e1 exp e2
  | Sequence (e1, e2, es) ->
    let seq b es =
      List.iter (fun e -> bprintf b ";\n%a" exp e) es in
    bprintf b "%a%a"
      exp e1
      seq (e2 :: es)
  | Record l -> bprintf b "{ %a}" fields l
  | Tuple es -> tuple b es
  | List es -> bprintf b "[%a]" (exp_sep_exps "; ") es
  | Function cs -> bprintf b "function%a\n" clauses cs
  | Comment s -> bprintf b "\n(* %s *)\n" s

and patterns b = sep_exps " " b

and pattern b = exp b

and sep_exp sep b = bprintf b "%s%a" sep exp

and sep_exp_parens sep b e =
  if needs_parens e
  then bprintf b "%s(%a)" sep exp e
  else bprintf b "%s%a" sep exp e

and sep_exps sep = iter (sep_exp sep)

and sep_exps_parens sep = iter (sep_exp_parens sep)

and exp_sep_exps sep b = function
  | [] -> ()
  | e :: es -> bprintf b "%a%a" exp e (sep_exps sep) es

and papply sep b e l = bprintf b "(%a%a)" exp e (sep_exps_parens sep) l

and tuple b = function
  | [] -> ()
  | e :: es -> bprintf b "(%a%a)" exp e (sep_exps ", ") es

and clause b = function
  | Clause (e1, e2) -> bprintf b "\n| %a -> %a" pattern e1 exp e2
  | Clause_when (e1, w, e2) ->
    bprintf b "\n| %a when %a -> %a" pattern e1 exp w exp e2

and clauses b = iter clause b

and field b (name, e) = bprintf b "%a = %a; " longident name exp e

and fields b = iter field b
;;

let rec structure_item b = function
  | Pstr_eval e -> bprintf b "\n%a;;\n" exp e
  | Pstr_value (_, []) -> ()
  | Pstr_value (Recursive, ((fid, args, e) :: rest)) ->
    bprintf b "\nlet rec %a%a = %a%a;;\n"
      longident fid patterns args exp e print_and rest
  | Pstr_value (Nonrecursive, ((fid, args, e) :: rest)) ->
    bprintf b "\nlet %a%a = %a%a;;\n"
      longident fid patterns args exp e print_and rest
  | Pstr_primitive (s, valdesc) ->
    bprintf b "\nexternal %s%a;;\n"
      s Pr_otype.value_description valdesc
  | Pstr_type tydecls ->
    bprintf b "\ntype %a;;\n" Pr_otype.type_declarations tydecls
  | Pstr_value (Default, _ :: _) -> assert false
    (* Default is not used by the OCaml parser. *)
  | Pstr_open mod_name -> bprintf b "\nopen %a;;\n" Pr_otype.blongident mod_name

and print_and b =
  List.iter
    (function (fid, args, e) ->
     bprintf b "\nand %a%a = %a" longident fid patterns args exp e)
;;

let structure = iter structure_item
;;

