(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Frdric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2008,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: code.mli,v 1.34 2008/02/11 23:05:59 weis Exp $ *)

(** {3 The code we generate.} *)

type lident = Longident.t;;
(** Lowercase ident. *)
type uident = Longident.t;;
(** Uppercase ident. *)
type iident = Longident.t;;
(** Infix ident. *)
type ident = Longident.t;;
(** Any kind of ident. *)

type iname = string;;
(** The name of an infix symbol. *)
type tname = string;;
(** The name of a type declared in the AST. *)

type exp =
   | Constant of ident
   | Var of lident
   | Match of exp * clause list
   | Try of exp * clause list
   | Generator of uident * exp
   | Apply of exp * exp list
   | Infix of exp * iident * exp
   | If of exp * exp * exp
   | Let of lident * pattern list * exp * exp
   | Letrec of lident * pattern list * exp * exp
   | Sequence of exp * exp * exp list
   | Record of (lident * exp) list
   | Tuple of exp list
   | List of exp list
   | Ocaml_pattern of Parsetree.pattern
   | Ocaml_expression of Parsetree.expression
   | Function of clause list
   | Comment of string

and pattern = exp

and clause =
  | Clause of pattern * exp
  | Clause_when of pattern * exp * exp;;

type structure_item =
   (** Evaluates an expression *)
   | Pstr_eval of exp
   (** Defines a set of possibly recursive functional values *)
   | Pstr_value of Asttypes.rec_flag * (lident * exp list * exp) list
   (** Defines one primitive *)
   | Pstr_primitive of string * Parsetree.value_description
   (** Defines one type *)
   | Pstr_type of (string * Parsetree.type_declaration) list
   (** Opens a module *)
   | Pstr_open of uident;;

type structure = structure_item list;;

(* Replace in a pattern all variables by underscore. *)
val underscored_ocaml_pattern : Parsetree.pattern -> Parsetree.pattern;;
val underscored_pattern : exp -> exp;;

(* Replace in a non-when clause lhs pattern all variables by underscore. *)
val underscored_clause : clause -> clause;;

val prefix_longident : string -> Longident.t -> Longident.t;;
(** To add a prefix string to a qualified (long) identifier. *)

(** To define a toplevel let rec definition in a module implementation. *)
val topletrec : (lident * pattern list * exp) list -> structure_item;;

(** Globally predefined identifiers. *)

val underscore : exp;;

val x : exp;;
val y : exp;;
val z : exp;;
val t : exp;;
val u : exp;;

val genr_args : string -> int -> pattern list;;
(** [genr_args root n] returns the list of patterns [root1; ...; rootn]. *)

val genr_numbered_args : string -> int -> (int * exp) list;;
(** [genr_numbered_args root n] returns [1,root1; ...; n,rootn]. *)

val make_Constant : string -> exp;;
(** Generate a constant expression from the name of an ident. *)

val make_Var : string -> exp;;
(** Generate a variable expression from the name of an ident. *)

(** Application macros. *)

val apply : lident -> exp list -> exp;;
val apply1 : lident -> exp -> exp;;
val apply2 : lident -> exp -> exp -> exp;;
val apply3 : lident -> exp -> exp -> exp -> exp;;
(** Apply an ident to its argument(s). *)

val infix_apply : exp -> iident -> exp -> exp;;
(** Apply an infix ident to its arguments. *)

val infix_name : exp -> iname -> exp -> exp;;
(** Apply a given infix name to its arguments. *)

(** Other useful auxilliaries. *)

val let0 : lident -> exp -> exp -> exp;;
val let1 : lident -> pattern -> exp -> exp -> exp;;
val pair : exp -> exp -> exp;;
val raise_error : string -> exp;;

(** Printing the code. *)

val clause : Buffer.t -> clause -> unit;;
val structure : Buffer.t -> structure -> unit;;
val structure_item : Buffer.t -> structure_item -> unit;;
val exp : Buffer.t -> exp -> unit;;
