(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Frdric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2008,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: genr_base.mli,v 1.41 2008/02/11 23:05:59 weis Exp $ *)

(** {3 Basic definitions for generation functions.} *)

open Code
open Parsetree
;;

(** {6 Maximum sharing handling.} *)
val set_sharable_target : unit -> unit
val get_sharable_target : unit -> bool
(** Do we generate maximum sharing ? *)
;;

val eq_function : exp -> exp -> exp
(** Returns the proper equality function, according to whether we have or have
    not maximally shared terms for the relational type at hand. *)
;;

(** {6 Handling function names.} *)

val construction_function_name : string -> string
(** Construction function name from a generator (string) name. *)
;;
val construction_function_ident : uident -> lident
(** Construction function ident from a generator ident. *)
;;
val construction_function : generator -> lident
(** Construction function ident from a generator. *)
;;

val prefix_ident : string -> lident -> lident
(** Adds a prefix to an identifier.
  [prefix_ident p id] is the identifier p_id *)
;;

val prefix_construction_function : string -> generator -> lident
(** Name of the generation functions for a given prefix name
   and a given generator:
   [prefix_generator_function p g] is the ident "prefix_name",
   where "prefix" is the contents of p,
   and "name" is the uncapitalized version of g. *)
val prefix_construction_function2 : string -> generator -> generator -> lident
(** Name of the generation functions for a given prefix name
   and the combination of two generators:
   [prefix_generator_function p g1 g2] is the ident "prefix_name1_name2",
   where "prefix" is the contents of p,
   and "name1", "name2" are the uncapitalized versions of g1, g2. *)
;;

(** {6 Application of generators.} *)

val construction : generator -> exp list -> exp
val construction0 : generator -> exp
val construction1 : generator -> exp -> exp
val construction2 : generator -> exp -> exp -> exp
(** The application of unary or binary construction function associated with the
    given generator. *)
;;

val generator : generator -> exp list -> exp
(** Apply the generator term constructor for the generator: in case of sharing,
    we properly call the sharing version of the generator. *)
;;
val generator0 : generator -> exp
val generator1 : generator -> exp -> exp
val generator2 : generator -> exp -> exp -> exp
(** Apply the generator term constructor for the generator: special cases for
    unary and binary generators. *)
;;

(** {6 Pattern construction for generators and functions.} *)

val genr_pattern : pattern -> pattern
(** Generation of a LHS of a pattern matching clause. *)
;;

val pattern : generator -> exp list -> exp
val pattern0 : generator -> exp
val pattern1 : generator -> exp -> exp
val pattern2 : generator -> exp -> exp -> exp
(** Similar to generator application for the pattern case. *)
;;

(** {6 Generation of functions for relational types.} *)

val genr_eq : (string * type_declaration) -> Code.structure_item
(** Generate an equality predicate primitive for the type name given:
    for type [t], we generate the primitive [eq_t] bound to [equal] if the
    values of the type are not maximally shared, or [eq] otherwise. *)
;;

val genr_projection : (string * type_declaration) ->
  core_type -> Code.structure_item
(** Generate a projection primitive for the (abbreviation) type name given. *)
;;

type function_definition = {
  fd_generator : generator;
  fd_relations : relations;
  fd_arity : int;
}
(** The data structure for the generation of construction functions. *)
;;

val clauses_of_side : relation_side -> clause -> clause -> clause list
;;
val clauses_left_or_right :
  relation_side -> clause -> clause -> clause list
;;
val clauses_both_as_left :
  relation_side -> clause -> clause -> clause list
;;
(** Functions to select a clause according to the side argument. *)

val opt : bool -> clause -> clause list -> clause list
;;
val lazy_opt : bool -> (unit -> clause) -> clause list -> clause list
;;
(** Functions to generate clauses on demand. *)

(* {6 Comparison functions.} *)

type binary_fun_exp = exp -> exp -> exp
(** The type of a function that takes two expression arguments and returns an
    expression that represents the application of the function to these
    arguments. *)
;;

type comparisons = binary_fun_exp * binary_fun_exp * binary_fun_exp
;;
(** The triple of a lesser, an equal, and a greater predicate. *)

val comparison_function : Longident.t option -> exp
(** A constant representing the name of the comparsion function corresponding
    to the indicated identifier. Returns the standard comparison function
    if no identifier is provided. *)
;;

val comparison_function_in_rels : relations -> exp
(** A constant representing the name of the comparsion function corresponding
    to the relations, which must include commutativity.
    Returns the comparison function as specified in the commutativity relation,
    or the standard comparison function if none is specified. *)
;;

(** {6 Comparison expressions.} *)

(*val default_comparisons : comparisons
(** The usual default infix comparisons. *)
;;*)

val standard_comparisons : Longident.t -> comparisons
(** The standard encodings of comparisons in terms of
   expression compare x y < (or = or >) 0. *)
;;

val genr_comparisons : lident option -> comparisons
(** Generation of a triple of comparison functions for a given
   commutative operator. *)
;;

(* Basic idents and functions. *)

val map : distributivity -> ('a -> 'b) -> 'a list -> 'b list
;;
val imap : distributivity -> Longident.t
;;
