(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Frdric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2008,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: genr_binary.ml,v 1.96 2008/02/11 23:05:59 weis Exp $ *)

(* We treat the case of binary generators.

    Let's [g] be such a generator; we generate a binary function (here binary
    means that the function has a pair as unique argument) that normalizes
    terms that are built with [g] as their head constructor. *)

open Parsetree
open Relation
open Code
open Genr_base
open Longident
;;

let insert_in = prefix_construction_function "insert"
and insert_inv_in = prefix_construction_function "insert_inv"
and delete_in = prefix_construction_function "delete"
and is_redex_in = prefix_construction_function "is_redex"
and return_in = prefix_construction_function "return"
;;

(* Application of the relevant generated function for the generator g: the
   relevant apply* is used, the proper name is generated and arity is taken
   into account. *)
let is_redex g x y = apply1 (is_redex_in g) (pair x y)
and return g x y = apply1 (return_in g) (pair x y)
and insert g = apply2 (insert_in g)
and insert_inv g = apply2 (insert_inv_in g)
and delete g = apply2 (delete_in g)
;;

(* Generation of an insertion function in a right comb for generator g. *)
let genr_insert fdef =

  (* [insert_g x u] inserts x in the right comb u *)

  let g = fdef.fd_generator
  and relations = fdef.fd_relations in

  (* Arguments of insert *)
  let args = [x; u] in

  (* Is generator g idempotent, commutative, or nilpotent ? *)
  let idempotent = has_idempotent relations
  and commutative = has_commutative relations
  and associative = has_associative relations in

  let _lt, eq, gt = genr_comparisons (Relation.commutative relations) in

  let nilpotent_gen_opt = nilpotent relations in
  let nilpotent_gen () =
    match nilpotent_gen_opt with
    | None -> assert false
    | Some g -> construction0 g in
  let nilpotent =
    match nilpotent_gen_opt with
    | None -> false
    | Some _ -> true in

  (* We match on u *)
  let clauses =
    (opt idempotent
      (Clause_when
         (pattern2 g y underscore, eq x y, u))
    (lazy_opt nilpotent (fun () ->
      (Clause_when (pattern2 g y z, eq x y,
		    return g (nilpotent_gen()) z)))
    (opt (associative && commutative)
      (Clause_when (pattern2 g y t, gt x y,
		    return g y (insert g x t)))
    (opt (associative && commutative)
       (Clause (pattern2 g underscore underscore, return g x u))
    (opt commutative
      (Clause_when (underscore, gt x u, return g u x))
    (opt (idempotent && commutative)
      (Clause_when (underscore, eq x u, u))
    (lazy_opt nilpotent (fun () ->
      Clause_when (underscore, eq x u, nilpotent_gen()))
    ([ Clause (underscore, return g x u) ])))))))) in

  insert_in g, args, Match (u, clauses)
;;

(* Generation of a deletion function in a right comb for generator g:
   Cancel a term and its opposite. *)
let genr_delete fdef e =

  (* [delete x u] tries to delete x in the right comb u,
     otherwise raises Not_found *)

  let g = fdef.fd_generator
  and relations = fdef.fd_relations in

  (* Arguments of delete *)
  let args = [x; u] in

  (* Is g commutative ? *)
  let commutative = has_commutative relations in

  let lt, _eq, _gt = genr_comparisons (Relation.commutative relations) in

  (* We match on u *)
  let clauses =
    (opt commutative
       (Clause_when
         (pattern2 g y underscore, lt x y,
          raise_error "Not_found"))
    (opt true
       (Clause_when (pattern2 g y t, eq_function x y,
                     construction2 g (construction0 e) t ))
    (opt commutative
       (Clause (pattern2 g y t, construction2 g y (delete g x t)))
    ([ Clause_when (underscore, eq_function u x, construction0 e);
       Clause (underscore, raise_error "Not_found"); ])))) in

  delete_in g, args, Match (u, clauses)
;;

(* Generation of an insertion function for inverse of terms in a right
   comb for generator g: we cancel the term and its opposite or inserts
   the inverse term argument in the comb. *)
let genr_insert_inv g ginv =

  (* [insert_inv x u] tries to delete x in the right comb u,
     otherwise inserts (opp x) in u *)

  let opp = construction1 ginv in

  (* Arguments of insert_inv *)
  let args = [x; u] in

  let clauses = [Clause (make_Constant "Not_found", insert g (opp x) u)] in
  let body = Try (delete g x u, clauses) in
  insert_inv_in g, args, body
;;

(* Generation of a return function for generator g:
   Tests if the term argument is a redex for [g] (due to rewrite rules):
   if so, calls the generation function again;
   otherwise builds a term with [g] as its head constructor. *)
let genr_return g =

  (* Generation of the return function. *)

  (* Arguments of return *)
  let args = [z] in
  let def = If (is_redex g x y, construction2 g x y, generator2 g x y) in
  let clauses = [Clause (pair x y, def)] in
  return_in g, args, Match (z, clauses)
;;

(* Generation of a construction function for a binary generator,
   according to its algebraic properties given as relation specifications
   in the definition of its associated constructor in the relation
   (private) type definition. *)
let genr_clauses_by_ap fdef =

  (* Generating the relevant clauses of the construction function for
     the generator given by [fdef], using the relations stated by the
     algebraic property [relation]. *)

  let c = fdef.fd_generator in

  function relation ->

  match relation.prel_desc with

  | Rewrite (pat, expr) -> [Genr_expression.genr_rule pat expr]

  | Commutative _ -> []

  | Idempotent s ->
    let left_clause =
      Clause_when
        (pair x (infix_name (pattern2 c y underscore) "as" z),
	 eq_function x y,
         z)
    and right_clause =
      Clause_when
        (pair (infix_name (pattern2 c underscore x) "as" z) y,
	 eq_function x y,
         z) in
    clauses_of_side s left_clause right_clause

  | Nilpotent (s, a) ->
    let left_clause =
      Clause_when
        (pair x (pattern2 c y z), eq_function x y,
         construction2 c (construction0 a) z)
    and right_clause =
      Clause_when
        (pair (pattern2 c z x) y, eq_function x y,
         construction2 c z (construction0 a)) in
    clauses_of_side s left_clause right_clause

  | Associative s ->
    let left_clause =
      Clause
        (pair x (pattern2 c y z),
         construction2 c (construction2 c x y) z)
    and right_clause =
      Clause
        (pair (pattern2 c x y) z,
         construction2 c x (construction2 c y z)) in
    clauses_left_or_right s left_clause right_clause

  (*| Absorbent (s, a) ->
    (* c (a, _) -> a. *)
    let left_clause =
      Clause (pair (pattern0 a) underscore, construction0 a)
    and right_clause =
      Clause (pair underscore (pattern0 a), construction0 a) in
    clauses_of_side s left_clause right_clause*)

  | Neutral (s, e) ->
    (* c (e, y) -> y. *)
    let left_clause = Clause (pair (pattern0 e) y, y)
    and right_clause = Clause (pair x (pattern0 e), x) in
    clauses_of_side s left_clause right_clause

  | Inverse (s, i, _) ->
    (* c (inv(x), x) -> e. when e = neutral of c. *)
    let left_clause = Clause (pair (pattern1 i x) y, insert_inv c x y)
    and right_clause =
      Clause (pair x (pattern1 i y), (*FIXME*)insert_inv c y x) in
    clauses_both_as_left s left_clause right_clause

  | Distributive (s, d, gopt, dflag) ->

    let e = match gopt with Some g -> g | None -> d in

    (* c(d[x1,..,xn], z) -> e[c(x1,z),..,c(xn,z)] if b = Dist_Direct
                            e[c(xn,z),..,c(x1,z)] if b = Dist_Inverse *)
    if Check.is_varyadic_gen d then (* e must be varyadic too *)
      let idist = Lident "dist" in
      let left_clause =
        Clause
          (pair (pattern1 d x) z,
           let1 idist y (construction2 c y z)
             (construction1 e (apply2 (imap dflag) (Constant idist) x)))
      and right_clause =
        Clause
          (pair z (pattern1 d x),
           let1 idist y (construction2 c z y)
             (construction1 e (apply2 (imap dflag) (Constant idist) x))) in
      clauses_of_side s left_clause right_clause

    (* c(d(x1,..,xn), z) -> e(c(x1,z),..,c(xn,z)) if b = Dist_Direct
                            e(c(xn,z),..,c(x1,z)) if b = Dist_Inverse *)
    else (* d and e must have the same arity *)
      begin match Check.arity_gen d with
	| 0 ->
	    let left_clause =
	      Clause (pair (pattern0 d) underscore, construction0 e)
	    and right_clause =
	      Clause (pair underscore (pattern0 d), construction0 e)
	    in
	      clauses_of_side s left_clause right_clause
	| n ->
	    let xs = genr_args "x" n in
	    let rhs dist = construction e (map dflag dist xs) in
	    let left_clause =
              let dist y = construction2 c y z in
		Clause (pair (pattern d xs) z, rhs dist)
	    and right_clause =
              let dist y = construction2 c z y in
		Clause (pair z (pattern d xs), rhs dist)
	    in
	      clauses_of_side s left_clause right_clause
      end

  | Absorbing (s, d) ->
    (* c (d (x, y), y) -> y. *)
    let left_clause =
      Clause_when
        (pair (pattern2 d x y) z, eq_function y z, y)
    and right_clause =
      Clause_when
        (pair x (pattern2 d y z), eq_function x y, x) in
    clauses_of_side s left_clause right_clause

  (* properties for unary generators *)
  | Division_by_Absorbent _ | Involutive -> assert false
  (* other properties *)
  | Structure_item _ | Precedence _ | Status _ -> []
;;

(* From a construction function definition, we generate the list of clauses
   that treats the relations of the generator. *)
let clauses_of_generator fdef =
  let relations = fdef.fd_relations in
  match relations.prels_desc with
  | Prels_none | Prels_commented _ -> []
  | Prels_begend rels ->
    List.flatten (List.map (genr_clauses_by_ap fdef) rels)
;;

(* Generation of a reducible patterns for a binary generator,
   according to its algebraic properties given as relation specifications
   in the definition of its associated constructor in the relation
   (private) type definition. *)

let genr_true_clauses_commutativity fdef =

  let id_true = make_Constant "true"
  and id_false = make_Constant "false"
  and u = underscore in

  let g = fdef.fd_generator in
  let relations = fdef.fd_relations in
  let _lt, _eq, gt = genr_comparisons (Relation.commutative relations)
  and associative = has_associative relations
  and commutative = has_commutative relations in

  if commutative then
    opt associative
      (Clause_when (pair x (pattern2 g y u), gt x y, id_true))
      (opt associative
         (Clause (pair u (pattern2 g u u), id_false))
         [Clause_when (pair x y, gt x y, id_true)])
  else []
;;

let genr_true_clauses_by_ap =

  let id_true = make_Constant "true"
  and u = underscore in

  let clause p = Clause (p, id_true)
  and clause_when p c = Clause_when (p, c, id_true) in

  let when_x_eq_y p = clause_when p (eq_function x y) in

  let clauses_left_or_right_pat s p1 p2 =
    clauses_left_or_right s (clause p1) (clause p2) in

  let clauses_of_side_pat s p1 p2 =
    clauses_of_side s (clause p1) (clause p2) in

  let rec underscores n = if n <= 0 then [] else u :: underscores (n-1) in

  fun fdef ->

    let g = fdef.fd_generator in
    let relations = fdef.fd_relations in
    (*let _lt, _eq, gt =
      genr_comparisons (Relation.commutative relations) in*)
    let associative = has_associative relations in

  fun relation ->

    match relation.prel_desc with

    | Rewrite (p, e) ->
	let clause =
	  match Genr_expression.genr_rule p e with
	  | Clause (p, _) -> Clause (underscored_pattern p, id_true)
	  | Clause_when (p, c, _) -> Clause_when (p, c, id_true)
	in [clause]

    | Commutative _ -> []
	(* opt associative
	  (clause_when (pair x (pattern2 g y u)) (gt x y))
	  [clause_when (pair x y) (gt x y)] *)

    | Idempotent s
    | Nilpotent (s, _) ->
	let left_clause = when_x_eq_y (pair x (pattern2 g y u))
	and right_clause = when_x_eq_y (pair (pattern2 g u x) y) in
	  (*when_x_eq_y (pair x y) ::*)
	  clauses_of_side s left_clause right_clause

    | Associative s ->
	let left_pat = pair u (pattern2 g u u)
	and right_pat = pair (pattern2 g u u) u in
	clauses_left_or_right_pat s left_pat right_pat

    | Neutral (s, a) ->
	let left_pat = pair (pattern0 a) u
	and right_pat = pair u (pattern0 a) in
	clauses_of_side_pat s left_pat right_pat

    | Inverse (s, i, _) ->
	let left_clause = when_x_eq_y (pair (pattern1 i x) y)
	and right_clause = when_x_eq_y (pair x (pattern1 i y)) in
	let clauses = clauses_of_side s left_clause right_clause in
	if associative then
	  let left_clause = when_x_eq_y
	      (pair (pattern1 i x) (pattern2 g y u))
	  and right_clause = when_x_eq_y
	      (pair x (pattern2 g (pattern1 i y) u))
	  in clauses @ clauses_of_side s left_clause right_clause
	else clauses

    | Distributive (s, d, _, _) ->
	if Check.is_varyadic_gen d then
	  let left_pat = pair (pattern1 d u) u
	  and right_pat = pair u (pattern1 d u) in
	    clauses_of_side_pat s left_pat right_pat
	else
	  let us = underscores (Check.arity_gen d) in
	  let left_pat = pair (pattern d us) u
	  and right_pat = pair u (pattern d us) in
	    clauses_of_side_pat s left_pat right_pat

    | Absorbing (s, d) ->
	let left_clause = when_x_eq_y (pair (pattern2 d u x) y)
	and right_clause = when_x_eq_y (pair x (pattern2 d y u))
	in clauses_of_side s left_clause right_clause

    (* properties for unary generators *)
    | Division_by_Absorbent _ | Involutive -> assert false
    (* other properties *)
    | Structure_item _ | Precedence _ | Status _ -> []
;;

(* From a construction function definition, we generate the list of
   clauses that treats the relations of the generator. *)
let true_clauses_of_generator fdef =
  let relations = fdef.fd_relations in
    match relations.prels_desc with
      | Prels_none | Prels_commented _ -> []
      | Prels_begend rels ->
	  List.flatten (List.map (genr_true_clauses_by_ap fdef) rels)
	  @ genr_true_clauses_commutativity fdef
;;

(* Generation of the predicate that tests if a given term with head
   constructor [g] can be a redex for some rewrite rule. If so, the
   [return_g] function should apply the construction function again;
   otherwise the [return_g] function directly returns a term with [g] as
   its head constructor (a term in [g] head normal form). *)
let genr_is_redex fdef =

  (* Generate the predicate to check whether a Rewrite rule can be
     applied to a given term. *)

  let g = fdef.fd_generator in

  (* Arguments of is_redex *)
  let args = [z] in

  (*let id_true = make_Constant "true" in

  let true_clause = function
    | Clause (l, _) -> Clause (underscored_pattern l, id_true)
    | Clause_when (l, c, _) -> Clause_when (l, c, id_true) in

  let clauses = List.map true_clause (clauses_of_generator fdef) in*)

  let clauses = true_clauses_of_generator fdef
  and default = [Clause (underscore, make_Constant "false")] in
  is_redex_in g, args, Match (z, clauses @ default)
;;

(* Generating the construction function for a binary generator.

   The (binary) generator is given by [fdef].

   We generate the clauses according to the relations recorded with the
   generator, and add a default insert clause. *)

let genr_construction_function fdef =

  let g = fdef.fd_generator in

  (* Arguments of [construction_function_name g] *)
  let args = [z] in

  let insert_fun =
    match inverse fdef.fd_relations with
    | Some (_si, i, _e) -> fun x y -> insert_inv g (construction1 i x) y
    | None -> insert g in

   let clauses =
     clauses_of_generator fdef @ [Clause (pair x y, insert_fun x y)] in

  construction_function g, args, Match (z, clauses)
;;

(* Generating the list of recursive definitions that defines the
construction function for generator g with the appropriate
auxilliaries (insert_*, delete_*, ...) functions associated to [g]. *)

let genr_function fdef =

  let g = fdef.fd_generator
  and relations = fdef.fd_relations in

  let construction_function = genr_construction_function fdef

  and insert_functions = [
    genr_is_redex fdef;
    genr_return g;
    genr_insert fdef;
  ]

  and delete_functions =
    match Relation.inverse relations, Relation.neutral relations with
    | Some (_si, i, Some e), _
    | Some (_si, i, None), Some (_, e) ->
        [genr_delete fdef e;
         genr_insert_inv g i]
    | _ -> [] in

  construction_function :: delete_functions @ insert_functions
;;
