(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Frdric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2008,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: pr_code.ml,v 1.11 2008/02/12 17:48:34 lowenthl Exp $ *)

open Format
open Code
;;

(* Printing the code. *)

type print_context =
   | Ctx_top
   | Ctx_list_elem
   | Ctx_fun
   | Ctx_fun_arg
   | Ctx_infix_arg
   | Ctx_if_guard
   | Ctx_let_def
   | Ctx_let_in
   | Ctx_match
   | Ctx_clause_when
   | Ctx_clause_right
   | Ctx_pattern
;;

(* Parentheses *)

let parens fmt = "(" ^^ fmt ^^ ")"
;;

let parens_match fmt = function
  | Ctx_top -> fmt
  | Ctx_list_elem -> fmt
  | Ctx_fun -> fmt
  | Ctx_fun_arg -> parens fmt
  | Ctx_infix_arg -> fmt
  | Ctx_if_guard -> fmt
  | Ctx_let_def -> fmt
  | Ctx_let_in -> fmt
  | Ctx_match -> fmt
  | Ctx_clause_when -> parens fmt
  | Ctx_clause_right -> parens fmt
  | Ctx_pattern -> parens fmt
;;

let parens_apply fmt = function
  | Ctx_top -> fmt
  | Ctx_list_elem -> fmt
  | Ctx_fun -> parens fmt
  | Ctx_fun_arg -> parens fmt
  | Ctx_infix_arg -> fmt
  | Ctx_if_guard -> fmt
  | Ctx_let_def -> fmt
  | Ctx_let_in -> fmt
  | Ctx_match -> fmt
  | Ctx_clause_when -> fmt
  | Ctx_clause_right -> fmt
  | Ctx_pattern -> fmt
;;

let parens_infix fmt = function
  | Ctx_top -> fmt
  | Ctx_list_elem -> parens fmt
  | Ctx_fun -> parens fmt
  | Ctx_fun_arg -> parens fmt
  | Ctx_infix_arg -> parens fmt
  | Ctx_if_guard -> fmt
  | Ctx_let_def -> fmt
  | Ctx_let_in -> fmt
  | Ctx_match -> fmt
  | Ctx_clause_when -> fmt
  | Ctx_clause_right -> fmt
  | Ctx_pattern -> fmt
;;

let parens_if fmt = function
  | Ctx_top -> fmt
  | Ctx_list_elem -> fmt
  | Ctx_fun -> parens fmt
  | Ctx_fun_arg -> parens fmt
  | Ctx_infix_arg -> parens fmt
  | Ctx_if_guard -> fmt
  | Ctx_let_def -> fmt
  | Ctx_let_in -> fmt
  | Ctx_match -> parens fmt
  | Ctx_clause_when -> parens fmt
  | Ctx_clause_right -> fmt
  | Ctx_pattern -> parens fmt
;;


let parens_let fmt = function
  | Ctx_top -> fmt
  | Ctx_list_elem -> fmt
  | Ctx_fun -> parens fmt
  | Ctx_fun_arg -> parens fmt
  | Ctx_infix_arg -> parens fmt
  | Ctx_if_guard -> fmt
  | Ctx_let_def -> fmt
  | Ctx_let_in -> fmt
  | Ctx_match -> parens fmt
  | Ctx_clause_when -> parens fmt
  | Ctx_clause_right -> parens fmt
  | Ctx_pattern -> parens fmt
;;

let parens_sequence fmt = function
  | Ctx_top -> fmt
  | Ctx_list_elem -> parens fmt
  | Ctx_fun -> parens fmt
  | Ctx_fun_arg -> parens fmt
  | Ctx_infix_arg -> parens fmt
  | Ctx_if_guard -> parens fmt
  | Ctx_let_def -> parens fmt
  | Ctx_let_in -> fmt
  | Ctx_match -> parens fmt
  | Ctx_clause_when -> parens fmt
  | Ctx_clause_right -> fmt
  | Ctx_pattern -> parens fmt
;;

(* Auxiliaries *)
let pr_via_buffer pr_buf f e =
  let b = Buffer.create 16 in
  pr_buf b e;
  pp_print_string f (Buffer.contents b)
;;

(* Printing expressions *)

let pr_prefix_exps prefix pr f =
  List.iter (fun e -> fprintf f "%s@ %a" prefix pr e)
;;

let pr_sep_exps sep pr f = function
  | [] -> ()
  | e :: es -> fprintf f "@[%a@]" pr e; pr_prefix_exps sep pr f es
;;

let pr_space_exps pr f = pr_prefix_exps "" pr f
;;

let rec pr_exp ctx f = function
  | Constant lid -> fprintf f "%a" pr_ident lid
  | Var lid -> fprintf f "%a" pr_ident lid
  | Ocaml_pattern p -> pr_ocaml_pattern f p
  | Ocaml_expression e -> pr_ocaml_expression f e
  | Match (e, cs) -> pr_match ctx f "match" e cs
  | Try (e, cs) -> pr_match ctx f "try" e cs
  | Generator (gid, e) when Pr_otype.is_infix_longident gid ->
     pr_infix_generator ctx f gid e
  | Generator (gid, e) -> pr_generator ctx f gid e
  | Apply (e, l) -> pr_apply ctx f e l
  | Infix (e1, lid, e2) -> pr_infix ctx f e1 lid e2
  | If (e1, e2, e3) -> pr_if ctx f e1 e2 e3
  | Let (name, args, e1, e2) -> pr_let "let" ctx f name args e1 e2
  | Letrec (name, args, e1, e2) -> pr_let "let rec" ctx f name args e1 e2
  | Sequence (e1, e2, es) -> pr_sequence ctx f e1 e2 es
  | Record fs -> pr_record ctx f fs
  | Tuple es -> pr_tuple ctx f es
  | List es -> pr_list ctx f es
  | Function cs -> pr_function ctx f cs
  | Comment s -> fprintf f "@[(* %s *)@]" s

and pr_ident = pr_via_buffer Pr_otype.infix_longident

and pr_idents f es = pr_space_exps pr_ident f es

and pr_ocaml_expression = pr_via_buffer Pr_otype.expression

and pr_ocaml_pattern = pr_via_buffer Pr_otype.pattern

and pr_match ctx f keyw e cs =
  let fmt_string = parens_match "@[<v 2>@[%s@ %a@ with@]%a@]" ctx in
  fprintf f fmt_string keyw pr_match_exp e pr_clauses cs

and pr_match_exp f e = pr_exp Ctx_match f e

and pr_clause f = function
  | Clause (p, e) -> fprintf f "@;| @[<2>%a ->@ %a@]"
      pr_pattern p pr_clause_right_side e
  | Clause_when (p, w, e) -> fprintf f "@;| @[<2>%a@ when %a ->@ %a@]"
      pr_pattern p pr_when w pr_let_in e

and pr_when f e = pr_exp Ctx_clause_when f e

and pr_clause_right_side f e = pr_exp Ctx_clause_right f e

and pr_clauses f cs = List.iter (pr_clause f) cs

and pr_infix_generator ctx f gid e =
  match e with
   | Tuple [e1; e2] -> pr_infix ctx f e1 gid e2
   | Constant _ | Var _ | Ocaml_pattern _
   | Ocaml_expression _ | Match _ | Try _
   | Generator _ | Apply _ | Infix _ | If _ | Let _
   | Letrec _ | Sequence _ | Record _    | Tuple _ | List _
   | Function _ | Comment _ -> pr_generator ctx f gid e

and pr_generator ctx f gid e =
  let fmt_string = parens_apply "%a %a" ctx in
  fprintf f fmt_string pr_ident gid pr_fun_arg e

and pr_apply ctx f e l =
  let fmt_string = parens_apply "%a%a" ctx in
  fprintf f fmt_string pr_fun_call e pr_fun_args l

and pr_fun_call f e = pr_exp Ctx_fun f e

and pr_fun_arg f e = pr_exp Ctx_fun_arg f e

and pr_fun_args f args = pr_space_exps pr_fun_arg f args

and pr_infix ctx f e1 lid e2 =
  let fmt_string = parens_infix "@[<1>%a@ %a@ %a@]" ctx in
  fprintf f fmt_string pr_infix_arg e1 pr_ident lid pr_infix_arg e2

and pr_infix_arg f e = pr_exp Ctx_infix_arg f e

and pr_if ctx f e1 e2 e3 =
  let fmt_string = parens_if "@[<hv 0>if @[%a@]@ then @[%a@]@ else @[%a@]@]" ctx in
  fprintf f fmt_string pr_if_guard e1 pr_if_branch e2 pr_if_branch e3

and pr_if_guard f e = pr_exp Ctx_if_guard f e

and pr_if_branch f e = pr_exp Ctx_let_in f e

and pr_let keyw ctx f name args e1 e2 =
  let fmt_string = parens_let "@[<v 0>@[<2>%s %a%a =@ @[<2>%a@] in@]@,@[%a@]@]" ctx in
  fprintf f fmt_string keyw pr_ident name pr_patterns args pr_let_def e1 pr_let_in e2

and pr_let_def f e = pr_exp Ctx_let_def f e

and pr_let_in f e = pr_exp Ctx_let_in f e

and pr_pattern f p = pr_exp Ctx_pattern f p

and pr_patterns f ps  = pr_space_exps pr_pattern f ps

and pr_sequence ctx f e1 e2 es =
  let fmt_string = parens_sequence "@[<v 0>%a@]" ctx in
  fprintf f fmt_string (pr_sep_exps ";" pr_list_elem) (e1 :: e2 :: es)

and pr_record _ f fds =
  fprintf f "@[{ %a }@]" pr_fields fds

and pr_fields f fds = pr_sep_exps ";" pr_field f fds

and pr_field f (name, e) = fprintf f "%a =@ %a" pr_ident name pr_list_elem e

and pr_tuple _ f es = fprintf f "@[(%a)@]" (pr_sep_exps "," pr_list_elem) es

and pr_list _ f es = fprintf f "@[<1>[%a]@]" (pr_sep_exps ";" pr_list_elem) es

and pr_list_elem f e = pr_exp Ctx_list_elem f e

and pr_function ctx f cs =
  let fmt_string = parens_match "@[<v 2>@[function@]%a@]" ctx in
  fprintf f fmt_string pr_clauses cs
;;

(* Printing structure items *)

let pr_value f rec_flag =
  let let_keyw = match rec_flag with
    | Asttypes.Nonrecursive | Asttypes.Default -> "let"
    | Asttypes.Recursive -> "let rec" in
  let value_step keyw (ident, args, e) =
    fprintf f  "@[%s %a%a =@;<1 2>@[<2>%a@]@]" keyw
      pr_ident ident pr_patterns args pr_let_def e in
  function
  | [] -> ()
  | v :: vs ->
    pp_open_vbox f 0;
    value_step let_keyw v;
    List.iter (fun v -> pp_print_break f 1 0; pp_print_break f 1 0; value_step "and" v) vs;
    pp_close_box f ()
;;

let rec pr_primitive f s valdesc =
  fprintf f "@[external %s%a@]" s pr_value_description valdesc

and pr_value_description = pr_via_buffer Pr_otype.value_description
;;

let structure_item f = function
  | Pstr_eval e -> pr_exp Ctx_top f e
  | Pstr_value (rec_flag, exps) -> pr_value f rec_flag exps
  | Pstr_primitive (s, valdesc) -> pr_primitive f s valdesc
  | Pstr_type _ -> ()
  | Pstr_open mod_name -> fprintf f "@[open %a@]" Pr_otype.longident mod_name
;;

let structure_sep f = fprintf f "@,;;@.@.";;

let structure b s =
  let f = formatter_of_buffer b in
  let structure_step item =
    structure_item f item;
    structure_sep f in
  List.iter structure_step s
;;

let exp b e =
  let ppf = formatter_of_buffer b in
  pr_exp Ctx_top ppf e;
  pp_print_flush ppf ()
;;
