(**********************************************************************)
(*                                                                    *)
(*                           Moca                                     *)
(*                                                                    *)
(*          Pierre Weis, INRIA Rocquencourt                           *)
(*          Frdric Blanqui, projet Protheo, INRIA Lorraine          *)
(*                                                                    *)
(*  Copyright 2005-2008,                                              *)
(*  Institut National de Recherche en Informatique et en Automatique. *)
(*  All rights reserved.                                              *)
(*                                                                    *)
(*  This file is distributed under the terms of the Q Public License. *)
(*                                                                    *)
(**********************************************************************)

(* $Id: pr_otype.ml,v 1.73 2008/02/11 23:11:03 weis Exp $ *)

(** {3 Printing Objective Caml type definitions.} *)

open Asttypes
open Parsetree
open Longident
open Printf
open Lexing
open Location
;;

let prefix_list sep f =
  let rec aux ob = function
    | [] -> ()
    | x :: l -> bprintf ob "%s%a%a" sep f x aux l in
  aux
;;

let list sep f ob = function
  | [] -> ()
  | x :: l -> bprintf ob "%a%a" f x (prefix_list sep f) l
;;

let suffix_list sep f =
  let rec aux ob = function
    | [] -> ()
    | x :: l -> bprintf ob "%a%s%a" f x sep aux l in
  aux
;;

let option f ob = function
  | Some x -> bprintf ob "%a" f x
  | None -> ()
;;

let constant ob = function
  | Const_int i ->
    if i < 0 then bprintf ob "(%i)" i else bprintf ob "%i" i
  | Const_char c -> bprintf ob "%C" c
  | Const_string s -> bprintf ob "%S" s
  | Const_float s -> bprintf ob "%s" s
  | Const_int32 li -> bprintf ob "%li" li
  | Const_int64 lli -> bprintf ob "%Li" lli
  | Const_nativeint ni -> bprintf ob "%ni" ni
;;

let plabel ob = function
  | "" -> ()
  | l -> bprintf ob ":%s" l
;;

let is_infix s =
  String.length s > 0 &&
  match s.[0] with
  | '!' | '$' | '%' | '&' | '*' | '+' | '-' | '.' | '/' | ':' | '<'
  | '=' | '>' | '?' | '@' | '^' | '|' | '~' -> true
  | _ -> false
;;

let rec is_infix_longident = function
  | Lident s -> is_infix s
  | Ldot (_mod, s) -> is_infix s
  | Lapply (_funct, id) -> is_infix_longident id
;;

let ident ob s =
  if is_infix s then bprintf ob "( %s )" s else bprintf ob "%s" s
;;

let rec longident ob = function
  | Lident id -> ident ob id
  | Ldot (t, id) -> bprintf ob "%a.%a" mod_longident t ident id
  | Lapply (t1, t2) -> bprintf ob "%a(%a)" mod_longident t1 mod_longident t2

and mod_ext_longident ob = longident ob

(* Should be restricted to a chain of M.M. ... .UIDENT *)
and mod_longident ob = longident ob
;;

(* To print a generator. *)
let generator ob g = longident ob g.pgen_desc;;

(* Printing without () arount infixes.
   Needed to print + in 1 + 2 for instance. *)
let infix_ident ob s = bprintf ob "%s" s;;

let rec infix_longident ob = function
  | Lident s -> infix_ident ob s
  | Ldot (t, s) -> bprintf ob "%a.%a" infix_longident t infix_ident s
  | Lapply (t1, t2) ->
    bprintf ob "%a(%a)" infix_longident t1 infix_longident t2
;;

let label_longident ob = function
  | Lident id -> ident ob id
  | Ldot (mod_lid, s) -> bprintf ob "%a.%a" mod_longident mod_lid ident s
  | Lapply (_, _) -> assert false
;;

(* Printing types. *)
let rec core_type ob x =
  match x.ptyp_desc with
  | Ptyp_any -> bprintf ob "_"
  | Ptyp_var (s) -> bprintf ob "'%s" s
  | Ptyp_arrow (_l, ct1, ct2) ->
    bprintf ob "(%a -> %a)" core_type ct1 core_type ct2
  | Ptyp_tuple [] -> ()
  | Ptyp_tuple cts -> bprintf ob "(%a)" core_type_list cts
  | Ptyp_constr (li, []) -> bprintf ob "%a" longident li
  | Ptyp_constr (li, [ct]) -> bprintf ob "%a %a" core_type ct longident li
  | Ptyp_constr (li, cts) ->
    bprintf ob "(%a) %a" (list ", " core_type) cts longident li
  | Ptyp_variant _
  | Ptyp_object _
  | Ptyp_class _
  | Ptyp_alias _
  | Ptyp_poly _ ->
    failwith "Printing of variant, classes and aliases not yet implemented."
  | Ptyp_parens ct -> bprintf ob "(%a)" core_type ct

and core_type_list ob cts =
  bprintf ob "%a" (list " * " core_type) cts

and constructor_type_arguments ob = function
  | [] -> ()
  | args -> bprintf ob " of %a" core_type_list args
;;

(* Printing patterns. *)
let rec pattern ob { ppat_desc = p } =
  match p with
  | Ppat_any -> bprintf ob "_"
  | Ppat_var s -> bprintf ob "%a" ident s
  | Ppat_alias (p, s) -> bprintf ob "(%a as %a)" pattern p ident s
  | Ppat_constant c -> constant ob c
  | Ppat_tuple ([] | [_]) -> assert false
  | Ppat_tuple pats -> bprintf ob "(%a)" (list ", " pattern) pats
  | Ppat_construct (id, None, _) -> longident ob id
  | Ppat_construct (id, Some { ppat_desc = Ppat_tuple [p1; p2]}, _)
      when is_infix_longident id ->
    bprintf ob "(%a %a %a)" pattern p1 infix_longident id pattern p2
  | Ppat_construct (id, Some p, _) ->
    bprintf ob "(%a %a)" longident id pattern p
  | Ppat_variant (lbl, None) -> bprintf ob "`%s" lbl
  | Ppat_variant (lbl, Some p) ->
    bprintf ob "`%s %a" lbl pattern p
  | Ppat_record idpats ->
    let f ob (id, pat) = bprintf ob "%a = %a" label_longident id pattern pat in
    bprintf ob "{%a}" (list "; " f) idpats
  | Ppat_array pats -> bprintf ob "[|%a|]" (list "; " pattern) pats
  | Ppat_or (p1, p2) -> bprintf ob "(%a | %a)" pattern p1 pattern p2
  | Ppat_constraint (p, ct) ->
    bprintf ob "(%a : %a)" pattern p core_type ct
  | Ppat_type id -> bprintf ob "#%a" longident id
  | Ppat_parens p -> bprintf ob "(%a)" pattern p
;;

let value_description ob x =
  bprintf ob " : %a" core_type x.pval_type;
  match x.pval_prim with
  | [] -> ()
  | l -> bprintf ob " =\n  %a" (list " " (fun ob -> bprintf ob "%S")) l
;;

let side ob = function
  | Right -> bprintf ob " right"
  | Left -> bprintf ob " left"
  | Both -> ()
;;

let side_if_left ob = function
  | Left -> bprintf ob " left"
  | Right
  | Both -> ()
;;

let find_text_in_file loc =
  let loc_start = loc.loc_start in
  let fname, _, _ = get_pos_info loc_start in
  let pos_cnum = loc_start.pos_cnum in
  let pos_enum = loc.loc_end.pos_cnum in
  let ic =
    try open_in fname with
    | _ ->
      failwith
        (Printf.sprintf
           "find_text_in_file: file ``%s'' not found or not readable."
           fname) in
  let len = pos_enum - pos_cnum in
  let buff = String.create len in
  try
    seek_in ic pos_cnum;
    really_input ic buff 0 len;
    buff with
  | End_of_file ->
    failwith
      (Printf.sprintf
         "find_text_in_file: file ``%s'' is too short."
         fname)
;;

let verbatim ob loc = bprintf ob "%s" (find_text_in_file loc);;

let verbatim_expression ob e = verbatim ob e.pexp_loc;;
let verbatim_pattern ob p = verbatim ob p.ppat_loc;;
let verbatim_structure_item ob si = verbatim ob si.pstr_loc;;

let private_flag ob = function
  | Public -> ()
  | Private -> bprintf ob " private"
;;

let mutable_flag ob = function
  | Immutable -> ()
  | Mutable -> bprintf ob "mutable "
;;

let rec_flag ob = function
  | Recursive -> bprintf ob " rec"
  | Nonrecursive
  | Default -> ()
;;

let ptype_param ob p = bprintf ob "'%s" p
;;

let ptype_params ob = function
  | [] -> ()
  | [param] -> bprintf ob "%a " ptype_param param
  | params -> bprintf ob "(%a) " (list ", " ptype_param) params
;;

let rec ptype ob (s, ty_decl) =
  bprintf ob "%a%s" ptype_params ty_decl.ptype_params s
;;

let rec type_declaration ob (s, ty_decl) =
  ptype ob (s, ty_decl);
  (* Skip variance *)
  (* Skip constraints *)
  ptype_kind ob
    (ty_decl.ptype_kind, ty_decl.ptype_private, ty_decl.ptype_manifest)

and type_declarations ob = list "\nand " type_declaration ob

and ptype_manifest ob = function
  | None -> ()
  | Some cty -> bprintf ob " = %a" core_type cty

and ptype_kind ob (kind, pflag, ctype_opt) =
  match kind, pflag, ctype_opt with
  | Ptype_abstract _rels, Public, None -> ()
  | Ptype_abstract rels, Private, None ->
    bprintf ob " = private\n%a" relations rels
  | Ptype_abstract rels, Public, Some cty ->
    bprintf ob " = %a  %a" core_type cty relations rels
  | Ptype_abstract rels, Private, Some cty ->
    bprintf ob " (*= private\n %a*)\n%a" core_type cty relations rels
  | Ptype_variant cstrs, pflag, None ->
    bprintf ob " =%a%a" private_flag pflag constructor_declarations cstrs
  | Ptype_variant cstrs, pflag, Some cty ->
    bprintf ob " = %a =%a%a"
      core_type cty private_flag pflag constructor_declarations cstrs
  | Ptype_record lbls, pflag, None ->
    bprintf ob " =%a\n%a" private_flag pflag label_declarations lbls
  | Ptype_record lbls, pflag, Some cty ->
    bprintf ob " = %a =%a\n%a"
      core_type cty private_flag pflag label_declarations lbls

and module_binding ob mb =
  match mb.pmod_desc with
  | Pmod_constraint (me, mty) ->
    bprintf ob ": %a = %a" module_type mty module_expr me
  | Pmod_functor (s, mty, me) ->
    bprintf ob "(%s : %a) %a" s module_type mty module_expr me
  | _ -> bprintf ob "= %a" module_expr mb

and module_type ob mt =
  match mt.pmty_desc with
  | Pmty_ident mty_ident -> bprintf ob "%a" longident mty_ident
  | Pmty_signature l -> bprintf ob "sig\n  %a\nend\n" signature l
  | Pmty_functor (s, mty1, mty2) ->
      bprintf ob "functor ( %s : %a ) ->\n  %a"
        s module_type mty1 module_type mty2
  | Pmty_with (mty, clist) ->
      bprintf ob "%a with %a " module_type mty constraints clist
  | Pmty_parens mty ->
      bprintf ob "(%a)" module_type mty

and module_expr ob me =
  match me.pmod_desc with
  | Pmod_ident uid ->
    bprintf ob "%a" longident uid
  | Pmod_structure stru ->
    bprintf ob "struct\n  %a\nend\n" structure stru
  | Pmod_functor (s, mty, me) ->
    bprintf ob "functor (%s : %a) ->\n  %a"
      s module_type mty module_expr me
  | Pmod_apply (me1, me2) ->
    bprintf ob "%a (%a)" module_expr me1 module_expr me2
  | Pmod_constraint (me, mty) ->
    bprintf ob "(%a : %a)"  module_expr me module_type mty
  | Pmod_parens me ->
    bprintf ob "(%a)"  module_expr me

and primitive_declaration ob = function
  | [] -> ()
  | s :: prdecls -> bprintf ob "%s %a" s primitive_declaration prdecls

and signature_item ob sigitem =
  match sigitem.psig_desc with
  | Psig_value (v, {pval_type = pv; pval_prim = l;}) ->
    (match l with
     | [] -> bprintf ob "val %a : %a" ident v core_type pv
     | _ ->
       bprintf ob "external %a : %a =\n  %a"
         ident v core_type pv primitive_declaration l)
  | Psig_type tdl ->
    bprintf ob "type %a\n" type_declarations tdl
  | Psig_exception (s, args) ->
    bprintf ob "exception %s %a" s constructor_type_arguments args
  | Psig_module (s, mdecl) ->
    bprintf ob "module %s %a" s module_decl mdecl
  | Psig_recmodule mrdecls ->
    bprintf ob "module rec %a" recmodules mrdecls
  | Psig_modtype (ty, modty) ->
    (match modty with
     | Pmodtype_abstract -> bprintf ob "module type %s" ty
     | Pmodtype_manifest mty ->
       bprintf ob "module type %s =\n  %a" ty module_type mty)
  | Psig_open mid -> bprintf ob "open %a" longident mid
  | Psig_include mty -> bprintf ob "include %a" module_type mty
  (* TODO : some more object-related things to print *)
  | Psig_class _cldescr ->
      failwith "Printing of Psig_class not yet implemented."
  | Psig_class_type _cltdecl ->
      failwith "Printing of Psig_class_type not yet implemented."

and module_decl ob mdecl =
  match mdecl.pmty_desc with
  | Pmty_functor (s, mty, md) ->
    bprintf ob "(%s : %a) %a" s module_type mty module_decl md
  | _ -> bprintf ob ": %a" module_type mdecl

and recmodule ob (id, mty) = bprintf ob "%s : %a" id module_type mty

and recmodules ob = function
  | [re] -> bprintf ob "%a" recmodule  re
  | recmod :: recmodecls ->
    bprintf ob "%a\n  and %a" recmodule recmod recmodules recmodecls
   (* According to parser.mly, module_rec_declarations cannot be empty! *)
  | _ -> assert false

and signature ob = suffix_list ";;\n" signature_item ob

and structure ob = suffix_list ";;\n" structure_item ob

and structure_item ob str_item =
  match str_item.pstr_desc with
  | Pstr_eval expr -> bprintf ob "%a\n" expression expr
  | Pstr_value (rf, pel) ->
    let binding ob (p, e) = bprintf ob "%a = %a" pattern p expression e in
    bprintf ob "let%a %a\n"
      rec_flag rf (list "\nand " binding) pel
  | Pstr_primitive (str, val_descr) ->
    bprintf ob "external %a : %a =\n%a\n"
      ident str
      core_type val_descr.pval_type
      primitive_declaration val_descr.pval_prim
  | Pstr_type s_ty_decls ->
    bprintf ob "%a\n" type_declarations s_ty_decls
  | Pstr_exception (str, args) ->
    bprintf ob "exception %s of %a\n" str constructor_type_arguments args
  | Pstr_exn_rebind (str, lid) ->
    bprintf ob "exception %s = %a\n" str longident lid
  | Pstr_module (str, me) ->
    bprintf ob "module %s =%a\n"str module_expr me
  | Pstr_recmodule mod_bds ->
    let module_binding ob (str, mt, me) =
      bprintf ob "%s : %a =%a\n" str module_type mt module_expr me in
    (match mod_bds with
    | mod_bd :: mod_bds ->
      bprintf ob "module rec : %a =%a\n"
        module_binding mod_bd
        (fun ob ->
           List.iter
             (fun mod_bd -> bprintf ob "and %a\n" module_binding mod_bd))
        mod_bds
    | _ -> assert false)
  | Pstr_modtype (ty, modty) ->
    bprintf ob "module type %s = %a\n" ty module_type modty
  | Pstr_open lid -> bprintf ob "open %a\n" longident lid
  | Pstr_include me -> bprintf ob "include %a\n" module_expr me
  | Pstr_class _ -> failwith "Printing of Pstr_class not yet implemented."
  | Pstr_class_type _ ->
      failwith "Printing of Pstr_class_type not yet implemented."

and constraints _ob _cl =
  failwith "Printing of constraints not yet implemented."

(* TODO : print class structures*)
and class_struct _ob _c =
  failwith "Printing of class_struct not yet implemented."

(* Printing expressions. *)
and expression ob e =
  match e.pexp_desc with
  | Pexp_parens e -> bprintf ob "(%a)" expression e
  | Pexp_begend e -> bprintf ob "begin %a end" expression e
  | Pexp_ident lid -> longident ob lid
  | Pexp_constant c -> constant ob c
  | Pexp_let (rf, pel, e) ->
    let definition ob (p, e) = bprintf ob "%a = %a" pattern p expression e in
    bprintf ob "let%a %a in\n%a"
      rec_flag rf (list "\nand " definition) pel expression e
  | Pexp_function (l, eopt, pel) ->
    let clause ob (p, e) = bprintf ob "| %a -> %a" pattern p expression e in
    bprintf ob "function %s%a\n%a"
      l (option expression) eopt (list "\n" clause) pel
  | Pexp_apply ({pexp_desc = Pexp_ident (Lident s)}, ["", arg1; "", arg2])
      when is_infix s ->
    bprintf ob "%a %a %a" expression arg1 infix_ident s expression arg2
  | Pexp_apply (e, les) ->
    let lbl_expression ob (lbl, e) =
      bprintf ob "%a%a" plabel lbl expression e in
    bprintf ob "%a %a" expression e (list " " lbl_expression) les
  | Pexp_match (e, pel) ->
    let f ob (p, e) = bprintf ob "| %a -> %a" pattern p expression e in
    bprintf ob "match %a with\n%a" expression e (list "\n" f) pel
  | Pexp_try (e, pel) ->
    let f ob (p, e) = bprintf ob "| %a -> %a" pattern p expression e in
    bprintf ob "try %a with%a" expression e (list "\n" f) pel
  | Pexp_tuple el ->
    bprintf ob "%a" (list ", " expression) el
  | Pexp_construct (lid, None, _) -> longident ob lid
  | Pexp_construct (id, Some { pexp_desc = Pexp_tuple [e1; e2]}, _)
      when is_infix_longident id ->
    bprintf ob "(%a %a %a)" expression e1 infix_longident id expression e2
  | Pexp_construct (lid, Some e, _) ->
    bprintf ob "%a (%a)" longident lid expression e
  | Pexp_variant (l, None) -> bprintf ob "`%s" l
  | Pexp_variant (l, Some e) -> bprintf ob "`%s %a" l expression e
  | Pexp_record (fs, with_expr) ->
    let pr_with_expr ob = function
      | None -> ()
      | Some e -> bprintf ob "(%a) with " expression e in
    bprintf ob "{%a%a}" pr_with_expr with_expr lbl_expr_list fs
  | Pexp_field (e, lid) ->
    bprintf ob "%a.%a" expression e label_longident lid
  | Pexp_setfield (r, lbl, e) ->
    bprintf ob "%a.%a <- %a" expression r label_longident lbl expression e
  | Pexp_array exprs ->
    bprintf ob "[|%a|]" (suffix_list "; " expression) exprs
  | Pexp_ifthenelse (cond, then_expr, None) ->
    bprintf ob "if %a then %a" expression cond expression then_expr
  | Pexp_ifthenelse (cond, then_expr, Some else_expr) ->
    bprintf ob "if %a then %a else %a"
      expression cond expression then_expr expression else_expr
  | Pexp_sequence (expr1, expr2) ->
    bprintf ob "begin %a; %a end" expression expr1 expression expr2
  | Pexp_while (expr1, expr2) ->
    bprintf ob "while %a do\n%a\ndone" expression expr1 expression expr2
  | Pexp_for (idx, expr1, expr2, dir_flag, expr3) ->
    let pr_dir_flag ob dir_flag =
      bprintf ob "%s" (if dir_flag = Upto then "to" else "downto") in
    bprintf ob "for %a = %a%a %a do\n%a\ndone"
      ident idx expression expr1 pr_dir_flag dir_flag expression expr2
      expression expr3
  | Pexp_constraint (expr, core_type_option1, core_type_option2) ->
    let pr_constraint ob = function
      | Some ty, None -> bprintf ob ": %a" core_type ty
      | Some ty1, Some ty2 ->
          bprintf ob ": %a :> %a" core_type ty1 core_type ty2
      | None, Some ty -> bprintf ob ":> %a" core_type ty
      | _ -> () in
    bprintf ob "(%a %a)"
      expression expr pr_constraint (core_type_option1, core_type_option2)
  | Pexp_when (e1, e2) ->
    bprintf ob "when %a -> %a" expression e1 expression e2
  | Pexp_send (e, s)  -> bprintf ob "%a # %s" expression e s
  | Pexp_new t -> bprintf ob "new (%a)" longident t
  | Pexp_setinstvar (s, e) -> bprintf ob "%s <- %a" s expression e
  | Pexp_override l -> bprintf ob "{< %a  >}" (suffix_list "; " over_field) l
  | Pexp_letmodule (s, m, e) ->
    bprintf ob "let module %s %a in %a" s module_binding m expression e
  | Pexp_assert e ->
      bprintf ob "assert (%a)" expression e
  | Pexp_assertfalse ->
      bprintf ob "assert false"
  | Pexp_lazy e ->
      bprintf ob "lazy (%a)" expression e
  | Pexp_poly (e, ct) ->
    (match ct with
     | Some pty -> bprintf ob ": %a = %a" core_type pty expression e
     | None -> bprintf ob "%a" expression e)
  | Pexp_object c ->
    bprintf ob "object %a end" class_struct c

and lbl_expr_list ob = (list "; " lbl_expr) ob
and lbl_expr ob (lbl, e) = bprintf ob "%a = %a" label_longident lbl expression e

and over_field ob (s, e) = bprintf ob "%s = %a" s expression e

(* Moca's relations. *)
and relation ob r =
  match r.prel_desc with
  | Rewrite (pat, expr) ->
    (match expr.pexp_desc with
     | Pexp_when (e1, e2) ->
       bprintf ob "rule %a when %a -> %a"
         pattern pat expression e1 expression e2
     | _ -> bprintf ob "rule %a -> %a" pattern pat expression expr)
  | Commutative None -> bprintf ob "commutative"
  | Commutative (Some f) -> bprintf ob "commutative (%a)" longident f
  | Associative sd -> bprintf ob "associative%a" side_if_left sd
  (*| Absorbent (sd, g) -> bprintf ob "absorbent%a (%a)" side sd generator g*)
  | Absorbing (sd, g) -> bprintf ob "absorbing%a (%a)" side sd generator g
  | Neutral (sd, g) -> bprintf ob "neutral%a (%a)" side sd generator g
(*  | Idempotent_Unary -> bprintf ob "idempotent unary"*)
(*  | Idempotent_AC -> bprintf ob "idempotent AC"*)
  | Idempotent sd -> bprintf ob "idempotent%a" side sd
(*  | Nilpotent_Unary g -> bprintf ob "unary nilpotent (%a)" generator g*)
(*  | Nilpotent_AC g -> bprintf ob "nilpotent AC (%a)" generator g*)
  | Nilpotent (sd, g) -> bprintf ob "nilpotent (%a, %a)" side sd generator g
  | Involutive -> bprintf ob "involutive"
  | Inverse (sd, g, None) -> bprintf ob "inverse%a (%a)" side sd generator g
  | Inverse (sd, g1, Some g2) ->
    bprintf ob "inverse%a (%a, %a)" side sd generator g1 generator g2
  | Distributive (sd, g, None, b) ->
    bprintf ob "distributive%a%a (%a)" inverse_opt b side sd generator g
  | Distributive (sd, g, Some h, b) ->
    bprintf ob "distributive%a%a (%a, %a)"
      inverse_opt b side sd generator g generator h
 (* | Distributive_Inverse (sd, g) ->
    bprintf ob "distributive inverse%a (%a)" side sd generator g*)
(*  | Distributive_Unary (g1, None) ->
    bprintf ob "unary distributive (%a)" generator g1
  | Distributive_Unary (g1, Some g2) ->
    bprintf ob "unary distributive (%a, %a)" generator g1 generator g2*)
 (* | Inverse_Neutral (g1, None) ->
    bprintf ob "inverse neutral (%a)" generator g1
  | Inverse_Neutral (g1, Some g2) ->
    bprintf ob "inverse neutral (%a, %a)" generator g1 generator g2*)
  | Division_by_Absorbent _ -> bprintf ob "division by absorbent"
  | Precedence _ -> ()
  | Status s -> bprintf ob "status %a\n" status s
  | Structure_item struct_item ->
    bprintf ob "%a;;\n" structure_item struct_item

and status ob = function
  | Lex -> bprintf ob "lex"
  | Mul -> bprintf ob "mul"

and inverse_opt ob = function
  | Dist_Inverse -> bprintf ob " inverse"
  | Dist_Direct -> ()

and relations ob rels =
  match rels.prels_desc with
  | Prels_none -> ()
  | Prels_commented {prels_desc = Prels_none | Prels_begend []; } -> ()
  | Prels_commented rels -> bprintf ob "\n(*%a*)" relations rels
  | Prels_begend rls ->
    bprintf ob "\nbegin\n%aend\n"
      (suffix_list "\n" relation) rls

and constructor_declaration ob (cname, ctys, rels, _loc) =
  bprintf ob "%a%a%a" ident cname constructor_type_arguments ctys relations rels

and constructor_declarations ob = prefix_list "\n| " constructor_declaration ob

and label_declaration ob (lbl, mflag, cty, rels, _loc) =
  bprintf ob "%a%a : %a%a" mutable_flag mflag ident lbl core_type cty relations rels

and label_declarations ob = suffix_list ";\n" label_declaration ob
;;

let implementation = structure
;;

let interface = signature
;;

let blongident = longident
;;

let rec longident ppf = function
  | Lident s ->
    if is_infix s
    then Format.fprintf ppf "( %s )" s
    else Format.fprintf ppf "%s" s
  | Ldot (t, s) -> Format.fprintf ppf "%a.%s" longident t s
  | Lapply (t1, t2) ->
    Format.fprintf ppf "%a(%a)" longident t1 longident t2
;;
